@@ -1,11 +1,12 @@
 use crate::error::{
     Error, InvalidFunctionArgument, ResponseError, UnknownResponseStatus,
-    UnsupportedResponseDataType,
+    UnsupportedQueryResultType,
 };
 use crate::response::*;
 use crate::selector::Selector;
 use crate::util::{validate_duration, RuleType, TargetState};
 use std::collections::HashMap;
+use url::Url;
 
 /// A helper enum that is passed to the [Client::new] function in
 /// order to avoid errors on unsupported connection schemes.
@@ -82,9 +83,9 @@ impl Client {
     ///
     ///     let s = sum(v, Some(Aggregate::By(&["cpu"])));
     ///
-    ///     let response = tokio_test::block_on( async { client.query(s, None, None).await });
+    ///     let response = tokio_test::block_on( async { client.query(s, None, None).await.unwrap() });
     ///
-    ///     assert!(response.is_ok());
+    ///     assert!(response.as_instant().is_some());
     ///
     ///     Ok(())
     /// }
@@ -94,7 +95,7 @@ impl Client {
         vector: impl std::fmt::Display,
         time: Option<i64>,
         timeout: Option<&str>,
-    ) -> Result<Response, Error> {
+    ) -> Result<QueryResultType, Error> {
         let url = format!("{}/query", self.base_url);
 
         let query = vector.to_string();
@@ -133,7 +134,7 @@ impl Client {
         end: i64,
         step: &str,
         timeout: Option<&str>,
-    ) -> Result<Response, Error> {
+    ) -> Result<QueryResultType, Error> {
         let url = format!("{}/query_range", self.base_url);
 
         validate_duration(step)?;
@@ -169,7 +170,7 @@ impl Client {
             .and_then(convert_query_response)
     }
 
-    /// Find time series by series selectors.
+    /// Find time series that match certain label sets ([Selector]s).
     ///
     /// ```rust
     /// use prometheus_http_query::{Client, Scheme, Selector, Error};
@@ -198,7 +199,7 @@ impl Client {
         selectors: &[Selector<'_>],
         start: Option<i64>,
         end: Option<i64>,
-    ) -> Result<Response, Error> {
+    ) -> Result<Vec<HashMap<String, String>>, Error> {
         let url = format!("{}/series", self.base_url);
 
         let mut params = vec![];
@@ -248,17 +249,10 @@ impl Client {
             .map_err(Error::Reqwest)?;
 
         check_response(response).await.and_then(move |r| {
-            let data = r["data"].as_array().unwrap();
-
-            let mut result = vec![];
-
-            for datum in data {
-                let metric: HashMap<String, String> =
-                    serde_json::from_value(datum.to_owned()).unwrap();
-                result.push(metric);
-            }
-
-            Ok(Response::Series(result))
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let result: Vec<HashMap<String, String>> =
+                serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(result)
         })
     }
 
@@ -271,7 +265,7 @@ impl Client {
     ///     let client = Client::new(Scheme::Http, "localhost", 9090);
     ///
     ///     // To retrieve a list of all labels:
-    ///     let response = tokio_test::block_on( async { client.labels(None, None, None).await });
+    ///     let response = tokio_test::block_on( async { client.label_names(None, None, None).await });
     ///
     ///     assert!(response.is_ok());
     ///
@@ -285,19 +279,19 @@ impl Client {
     ///
     ///     let set = Some(vec![s1, s2]);
     ///
-    ///     let response = tokio_test::block_on( async { client.labels(set, None, None).await });
+    ///     let response = tokio_test::block_on( async { client.label_names(set, None, None).await });
     ///
     ///     assert!(response.is_ok());
     ///
     ///     Ok(())
     /// }
     /// ```
-    pub async fn labels(
+    pub async fn label_names(
         &self,
         selectors: Option<Vec<Selector<'_>>>,
         start: Option<i64>,
         end: Option<i64>,
-    ) -> Result<Response, Error> {
+    ) -> Result<Vec<String>, Error> {
         let url = format!("{}/labels", self.base_url);
 
         let mut params = vec![];
@@ -344,15 +338,9 @@ impl Client {
             .map_err(Error::Reqwest)?;
 
         check_response(response).await.and_then(move |r| {
-            let data = r["data"].as_array().unwrap();
-
-            let mut result = vec![];
-
-            for datum in data {
-                result.push(datum.as_str().unwrap().to_owned());
-            }
-
-            Ok(Response::LabelNames(result))
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let result: Vec<String> = serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(result)
         })
     }
 
@@ -388,7 +376,7 @@ impl Client {
         selectors: Option<Vec<Selector<'_>>>,
         start: Option<i64>,
         end: Option<i64>,
-    ) -> Result<Response, Error> {
+    ) -> Result<Vec<String>, Error> {
         let url = format!("{}/label/{}/values", self.base_url, label);
 
         let mut params = vec![];
@@ -435,15 +423,9 @@ impl Client {
             .map_err(Error::Reqwest)?;
 
         check_response(response).await.and_then(move |r| {
-            let data = r["data"].as_array().unwrap();
-
-            let mut result = vec![];
-
-            for datum in data {
-                result.push(datum.as_str().unwrap().to_owned());
-            }
-
-            Ok(Response::LabelValues(result))
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let result = serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(result)
         })
     }
 
@@ -468,7 +450,7 @@ impl Client {
     ///     Ok(())
     /// }
     /// ```
-    pub async fn targets(&self, state: Option<TargetState>) -> Result<Response, Error> {
+    pub async fn targets(&self, state: Option<TargetState>) -> Result<Targets, Error> {
         let url = format!("{}/targets", self.base_url);
 
         let mut params = vec![];
@@ -490,9 +472,9 @@ impl Client {
             .map_err(Error::Reqwest)?;
 
         check_response(response).await.and_then(move |r| {
-            let raw_targets = r["data"].to_owned();
-            let targets: Targets = serde_json::from_value(raw_targets).unwrap();
-            Ok(Response::Targets(targets))
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let targets: Targets = serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(targets)
         })
     }
 
@@ -517,7 +499,7 @@ impl Client {
     ///     Ok(())
     /// }
     /// ```
-    pub async fn rules(&self, rule_type: Option<RuleType>) -> Result<Response, Error> {
+    pub async fn rules(&self, rule_type: Option<RuleType>) -> Result<Vec<RuleGroup>, Error> {
         let url = format!("{}/rules", self.base_url);
 
         let mut params = vec![];
@@ -539,19 +521,19 @@ impl Client {
             .map_err(Error::Reqwest)?;
 
         check_response(response).await.and_then(move |r| {
-            let groups = r["data"].as_object().unwrap()["groups"]
-                .as_array()
+            let groups = r
+                .get("data")
+                .ok_or(Error::MissingField)?
+                .as_object()
                 .unwrap()
+                .get("groups")
+                .ok_or(Error::MissingField)?
                 .to_owned();
 
-            let mut result = vec![];
+            let result: Vec<RuleGroup> =
+                serde_json::from_value(groups).map_err(Error::ResponseParse)?;
 
-            for group in groups {
-                let g: Group = serde_json::from_value(group).unwrap();
-                result.push(g);
-            }
-
-            Ok(Response::Rules(result))
+            Ok(result)
         })
     }
 
@@ -571,7 +553,7 @@ impl Client {
     ///     Ok(())
     /// }
     /// ```
-    pub async fn alerts(&self) -> Result<Response, Error> {
+    pub async fn alerts(&self) -> Result<Vec<Alert>, Error> {
         let url = format!("{}/alerts", self.base_url);
 
         let response = self
@@ -584,19 +566,280 @@ impl Client {
             .map_err(Error::Reqwest)?;
 
         check_response(response).await.and_then(move |r| {
-            let alerts = r["data"].as_object().unwrap()["alerts"]
-                .as_array()
+            let alerts = r
+                .get("data")
+                .ok_or(Error::MissingField)?
+                .as_object()
                 .unwrap()
+                .get("alerts")
+                .ok_or(Error::MissingField)?
                 .to_owned();
 
-            let mut result = vec![];
+            let result: Vec<Alert> =
+                serde_json::from_value(alerts).map_err(Error::ResponseParse)?;
 
-            for alert in alerts {
-                let a: Alert = serde_json::from_value(alert).unwrap();
-                result.push(a);
+            Ok(result)
+        })
+    }
+
+    /// Retrieve a list of flags that Prometheus was configured with.
+    ///
+    /// ```rust
+    /// use prometheus_http_query::{Client, Scheme, Error};
+    /// use std::convert::TryInto;
+    ///
+    /// fn main() -> Result<(), Error> {
+    ///     let client = Client::new(Scheme::Http, "localhost", 9090);
+    ///
+    ///     let response = tokio_test::block_on( async { client.flags().await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
+    pub async fn flags(&self) -> Result<HashMap<String, String>, Error> {
+        let url = format!("{}/status/flags", self.base_url);
+
+        let response = self
+            .client
+            .get(&url)
+            .send()
+            .await
+            .map_err(Error::Reqwest)?
+            .error_for_status()
+            .map_err(Error::Reqwest)?;
+
+        check_response(response).await.and_then(move |r| {
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let flags: HashMap<String, String> =
+                serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(flags)
+        })
+    }
+
+    /// Query the current state of alertmanager discovery.
+    ///
+    /// ```rust
+    /// use prometheus_http_query::{Client, Scheme, Error, TargetState};
+    /// use std::convert::TryInto;
+    ///
+    /// fn main() -> Result<(), Error> {
+    ///     let client = Client::new(Scheme::Http, "localhost", 9090);
+    ///
+    ///     let response = tokio_test::block_on( async { client.alertmanagers().await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
+    pub async fn alertmanagers(&self) -> Result<Alertmanagers, Error> {
+        let url = format!("{}/alertmanagers", self.base_url);
+
+        let response = self
+            .client
+            .get(&url)
+            .send()
+            .await
+            .map_err(Error::Reqwest)?
+            .error_for_status()
+            .map_err(Error::Reqwest)?;
+
+        check_response(response).await.and_then(move |r| {
+            let data = r
+                .get("data")
+                .ok_or(Error::MissingField)?
+                .as_object()
+                .unwrap();
+
+            let mut active: Vec<Url> = vec![];
+
+            let items = data
+                .get("activeAlertmanagers")
+                .ok_or(Error::MissingField)?
+                .as_array()
+                .unwrap();
+
+            for item in items {
+                let raw_url = item
+                    .get("url")
+                    .ok_or(Error::MissingField)?
+                    .as_str()
+                    .unwrap();
+
+                let url = Url::parse(raw_url).map_err(Error::UrlParse)?;
+
+                active.push(url);
             }
 
-            Ok(Response::Alerts(result))
+            let mut dropped: Vec<Url> = vec![];
+
+            let items = data
+                .get("droppedAlertmanagers")
+                .ok_or(Error::MissingField)?
+                .as_array()
+                .unwrap();
+
+            for item in items {
+                let raw_url = item
+                    .get("url")
+                    .ok_or(Error::MissingField)?
+                    .as_str()
+                    .unwrap();
+
+                let url = Url::parse(raw_url).map_err(Error::UrlParse)?;
+
+                dropped.push(url);
+            }
+
+            let result = Alertmanagers { active, dropped };
+
+            Ok(result)
+        })
+    }
+
+    /// Retrieve metadata about metrics that are currently scraped from targets, along with target information.
+    ///
+    /// ```rust
+    /// use prometheus_http_query::{Client, Scheme, Error, Selector};
+    /// use std::convert::TryInto;
+    ///
+    /// fn main() -> Result<(), Error> {
+    ///     let client = Client::new(Scheme::Http, "localhost", 9090);
+    ///
+    ///     // Retrieve metadata for a specific metric from all targets.
+    ///     let response = tokio_test::block_on( async { client.target_metadata(Some("go_routines"), None, None).await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     // Retrieve metric metadata from specific targets.
+    ///     let s = Selector::new().with("job", "prometheus");
+    ///
+    ///     let response = tokio_test::block_on( async { client.target_metadata(None, Some(&s), None).await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     // Retrieve metadata for a specific metric from targets that match a specific label set.
+    ///     let s = Selector::new().with("job", "node");
+    ///
+    ///     let response = tokio_test::block_on( async { client.target_metadata(Some("node_cpu_seconds_total"), Some(&s), None).await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
+    pub async fn target_metadata(
+        &self,
+        metric: Option<&str>,
+        match_target: Option<&Selector<'_>>,
+        limit: Option<usize>,
+    ) -> Result<Vec<TargetMetadata>, Error> {
+        let url = format!("{}/targets/metadata", self.base_url);
+
+        let mut params = vec![];
+
+        let metric = metric.map(|s| s.to_string());
+
+        if let Some(m) = &metric {
+            params.push(("metric", m.as_str()))
+        }
+
+        let match_target = match_target.map(|s| s.to_string());
+
+        if let Some(m) = &match_target {
+            params.push(("match_target", m.as_str()))
+        }
+
+        let limit = limit.map(|s| s.to_string());
+
+        if let Some(l) = &limit {
+            params.push(("limit", l.as_str()))
+        }
+
+        let response = self
+            .client
+            .get(&url)
+            .query(params.as_slice())
+            .send()
+            .await
+            .map_err(Error::Reqwest)?
+            .error_for_status()
+            .map_err(Error::Reqwest)?;
+
+        check_response(response).await.and_then(move |r| {
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let result: Vec<TargetMetadata> =
+                serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(result)
+        })
+    }
+
+    /// Retrieve metadata about metrics that are currently scraped from targets.
+    ///
+    /// ```rust
+    /// use prometheus_http_query::{Client, Scheme, Error};
+    /// use std::convert::TryInto;
+    ///
+    /// fn main() -> Result<(), Error> {
+    ///     let client = Client::new(Scheme::Http, "localhost", 9090);
+    ///
+    ///     // Retrieve metadata for a all metrics.
+    ///     let response = tokio_test::block_on( async { client.metric_metadata(None, None).await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     // Limit the number of returned metrics
+    ///     let response = tokio_test::block_on( async { client.metric_metadata(None, Some(10)).await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     // Retrieve metadata of a specific metric.
+    ///     let response = tokio_test::block_on( async { client.metric_metadata(Some("go_routines"), None).await });
+    ///
+    ///     assert!(response.is_ok());
+    ///
+    ///     Ok(())
+    /// }
+    /// ```
+    pub async fn metric_metadata(
+        &self,
+        metric: Option<&str>,
+        limit: Option<usize>,
+    ) -> Result<HashMap<String, Vec<MetricMetadata>>, Error> {
+        let url = format!("{}/metadata", self.base_url);
+
+        let mut params = vec![];
+
+        let metric = metric.map(|s| s.to_string());
+
+        if let Some(m) = &metric {
+            params.push(("metric", m.as_str()))
+        }
+
+        let limit = limit.map(|s| s.to_string());
+
+        if let Some(l) = &limit {
+            params.push(("limit", l.as_str()))
+        }
+
+        let response = self
+            .client
+            .get(&url)
+            .query(params.as_slice())
+            .send()
+            .await
+            .map_err(Error::Reqwest)?
+            .error_for_status()
+            .map_err(Error::Reqwest)?;
+
+        check_response(response).await.and_then(move |r| {
+            let data = r.get("data").ok_or(Error::MissingField)?.to_owned();
+            let result: HashMap<String, Vec<MetricMetadata>> =
+                serde_json::from_value(data).map_err(Error::ResponseParse)?;
+
+            Ok(result)
         })
     }
 }
@@ -611,14 +854,31 @@ async fn check_response(
         .await
         .map_err(Error::Reqwest)?;
 
-    let status = map["status"].as_str().unwrap();
+    let status = map
+        .get("status")
+        .ok_or(Error::MissingField)?
+        .as_str()
+        .unwrap();
 
     match status {
         "success" => Ok(map),
-        "error" => Err(Error::ResponseError(ResponseError {
-            kind: map["errorType"].as_str().unwrap().to_string(),
-            message: map["error"].as_str().unwrap().to_string(),
-        })),
+        "error" => {
+            let kind = map
+                .get("errorType")
+                .ok_or(Error::MissingField)?
+                .as_str()
+                .unwrap()
+                .to_string();
+
+            let message = map
+                .get("error")
+                .ok_or(Error::MissingField)?
+                .as_str()
+                .unwrap()
+                .to_string();
+
+            Err(Error::ResponseError(ResponseError { kind, message }))
+        }
         _ => Err(Error::UnknownResponseStatus(UnknownResponseStatus(
             status.to_string(),
         ))),
@@ -627,34 +887,39 @@ async fn check_response(
 
 // Parses the API response from a map to a Response enum that
 // encapsulates a vector of samples of type "vector" or "matrix"
-fn convert_query_response(response: HashMap<String, serde_json::Value>) -> Result<Response, Error> {
-    let data_obj = response["data"].as_object().unwrap();
-    let data_type = data_obj["resultType"].as_str().unwrap();
-    let data = data_obj["result"].as_array().unwrap().to_owned();
+fn convert_query_response(
+    response: HashMap<String, serde_json::Value>,
+) -> Result<QueryResultType, Error> {
+    let data_obj = response
+        .get("data")
+        .ok_or(Error::MissingField)?
+        .as_object()
+        .unwrap();
+
+    let data_type = data_obj
+        .get("resultType")
+        .ok_or(Error::MissingField)?
+        .as_str()
+        .unwrap();
+
+    let data = data_obj
+        .get("result")
+        .ok_or(Error::MissingField)?
+        .to_owned();
 
     match data_type {
         "vector" => {
-            let mut result: Vec<Vector> = vec![];
-
-            for datum in data {
-                let vector: Vector = serde_json::from_value(datum).unwrap();
-                result.push(vector);
-            }
-
-            Ok(Response::Vector(result))
+            let result: Vec<InstantVector> =
+                serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(QueryResultType::Vector(result))
         }
         "matrix" => {
-            let mut result: Vec<Matrix> = vec![];
-
-            for datum in data {
-                let matrix: Matrix = serde_json::from_value(datum).unwrap();
-                result.push(matrix);
-            }
-
-            Ok(Response::Matrix(result))
+            let result: Vec<RangeVector> =
+                serde_json::from_value(data).map_err(Error::ResponseParse)?;
+            Ok(QueryResultType::Matrix(result))
         }
-        _ => Err(Error::UnsupportedResponseDataType(
-            UnsupportedResponseDataType(data_type.to_string()),
+        _ => Err(Error::UnsupportedQueryResultType(
+            UnsupportedQueryResultType(data_type.to_string()),
         )),
     }
 }
